# Kohonen map for the offline phase
# -------------------------------------------------------------------------------------
kohonen.offline.global <- function(offline.dataset,offline.classes,num.it,init.n,final.n,grid.d,tr.mode,min.ex){
  
  cat("\nOffline phase - building maps!")
  
  # Parameters of the algorithm
  n0 <- init.n # initial learning rate
  n1 <- final.n # final learnin rate
  # Learning rate will decline linearly from n0 to n1 over num.iterations updates 
  num.iterations <- num.it # maximum number of iteractions. 
  num.classes <- ncol(offline.classes) # number of initial classes. 
  train.mode <- tr.mode # training mode
  min.ex <- min.ex # minimum number of instances in a neuron
  n.neurons <- grid.d # grid dimension = n.neurons x n.neurons
  
  # Compute initial class probabilities and class totals
  r <- computeInitialClassProbabilitiesTotals(offline.classes)
  class.totals <- r$class.totals
  class.probabilities <- r$class.probabilities
  
  # Store label cardinality of the dataset
  z <- computeLabelCardinality(offline.classes)
  
  # Will store the amount of instances in each neuron
  instances.per.neuron <- vector()
  
  # Will store all properties computed here
  result <- list()

  # Name class totals and class probabilities columns and rows
  rownames(class.probabilities) <- paste("C",c(1:ncol(offline.classes)),sep="")
  colnames(class.probabilities) <- paste("C",c(1:ncol(offline.classes)),sep="")
  rownames(class.totals) <- paste("C",c(1:ncol(offline.classes)),sep="")
  colnames(class.totals) <- paste("C",c(1:ncol(offline.classes)),sep="")

  # Build som map with all data
  mygrid <- somgrid(n.neurons,n.neurons, topo=c("hexagonal"), neighbourhood.fct = c("gaussian"))
  som.map <- som(offline.dataset, grid=mygrid, rlen = num.iterations, alpha = c(n0, n1), dist.fcts=c("euclidean"), 
                 mode = c(train.mode), maxNA.fraction = 0L, cores=-1)#, keep.data = FALSE)
  
  # Store the som map and micro-clusters
  # [[1]] <- som map
  # [[2]] <- corresponding micro-clusters
  som.micro.clusters <- lapply(1, function(x) as.list(rep(0,2)))
  
  # Compute micro-clusters properties and store them
  resultMC <- computeMicroClusters(som.map,offline.classes,min.ex)
  
  # Get the average output of each neuron based on the mapped instances to them
  average.output.som.map <- getAverageNeuronOutputs(resultMC$som.map,length(resultMC$micro.clusters))
  
  # Calculates conditional probabilitiy thresholds for a class
  # We will consider a p(x|y_j) calculated from the training data (average.output.som.map) for the class for each neuron
  # That means we are considering the outputs of a neuron to be average.output.som.map to calculate the threshold
  # Each neuron will have a threshold
  resultMC$micro.clusters <- getCondProbabilitiesNeurons(resultMC$micro.clusters,class.probabilities,average.output.som.map)
  
  # Compute radius factor for each neuron
  #resultMC$micro.clusters <- computeRadiusFactorMC(resultMC$micro.clusters,resultMC$som.map,average.output.som.map,offline.dataset)
  
  # Store the trained SOM
  som.micro.clusters[[1]][[1]] <- resultMC$som.map
  
  # Store the corresponding micro-clusters
  som.micro.clusters[[1]][[2]] <-resultMC$micro.clusters

  # Store results
  result$som.micro.clusters <- som.micro.clusters
  result$z <- z
  result$class.probabilities <- class.probabilities
  result$class.totals <- class.totals
  result$total.instances <- nrow(offline.dataset)
  result$NP <- 0
  
  # Based on the minimum number of instances per neuron,
  # we will calculate the minimum number of instances required to create a new neuron (min.instances.neuron),
  # and the short-term memory storage limit threshold (theta)
  # First index contains the neurons not used
  if(as.numeric(names(table(resultMC$som.map$unit.classif)))[1] == 0){
    instances.per.neuron <- c(instances.per.neuron,table(resultMC$som.map$unit.classif)[-1])
  }
  else{
    instances.per.neuron <- c(instances.per.neuron,table(resultMC$som.map$unit.classif))
  }
  result$min.instances.neuron <- min(instances.per.neuron)
  result$theta <- n.neurons.grid*n.neurons.grid*result$min.instances.neuron
  
  # Will store total number of examples to consider when calculating probabilities for NP
  # When the ND procedure is called, if it creates new neurons, we will start to count the
  # total number of instances from this point (time.stamp - first.timestamp.unclassified)
  # first.timestamp.unclassified is the first time stamp of the examples in the short term memory
  # that were never explained by the current model
  total.instances.NP <- vector()
  result$total.instances.NP <- total.instances.NP
  
  # Will store the time stamp when a micro-cluster is created to represent a NP
  novel.patterns.time.stamp <- vector()
  result$novel.patterns.time.stamp <- novel.patterns.time.stamp

  return(result)
}
# -------------------------------------------------------------------------------------


# Kohonen map for the online phase considering bayes and label cardinality z
# -------------------------------------------------------------------------------------
kohonen.online.bayes.ND <- function(mapping,online.dataset,init.n,final.n,tr.mode,num.it,
                                    om,th,min.ex,window.stm.check,novel.classes,n.neurons.grid,
                                    update.model.info,num.offline.instances){
  
  #plot(1, type="n", xlab="", ylab="", xlim=c(0, 1), ylim=c(0, 1), main=paste(n.neurons.grid," neuronios ",sep=""))
  
  # Time stamp
  time.stamp <- 0
  
  # Learning rate
  n0 <- init.n
  n1 <- final.n
  
  # Training mode of the SOM grid
  train.mode <- tr.mode
  
  # Num iterations for Kohonen algorithm
  num.iterations <- num.it
  
  # Will store results
  results <- list()
  
  # Window size to forget outdated examples
  omega <- om
  
  # Short-term memory storage limit threshold
  theta <- th
  
  # Mininum number of examples for a valid cluster
  min.examples.cluster <- min.ex
  
  # Time stamp of the incoming examples
  time.stamp <- 0
  
  # Controls last time ND procedure was executed
  last.check <- 0
  window.check <- 0
  
  # Will store indexes of explained and unknown examples
  indexes.explained <- vector()
  indexes.unknown.classified <- vector()
  indexes.unknown.unclassified <- vector()
  # This is for evaluation purposes -> Compute unknown rate per window
  all.unclassified.indexes <- vector() 
  # This is for evaluation purposes -> Time stamp when an instance is removed from short-term memory
  time.stamp.stm.removed <- vector()
  
  # Short term memory
  short.term.memory.classified <- 0
  short.term.memory.unclassified <- 0
  
  # Number of examples in the stream
  num.examples <- nrow(online.dataset)
  
  # Prediction for the known classes
  predictions <- matrix(nrow=0, ncol=nrow(mapping$class.probabilities), data=0)
  colnames(predictions) <- paste("C",c(1:ncol(predictions)),sep="")
  
  # Number of initial classes
  initial.number.classes <- nrow(mapping$class.probabilities)
  
  # Analyse examples one by one
  cat("\nOnline phase!\n")
  step = 1
  for(inst in seq(from=1, to=num.examples, by=step)){
    
    matrix.examples <- 0
    if(inst+step > num.examples){
      matrix.examples <- online.dataset[inst:num.examples,,drop=F]
      rownames(matrix.examples) <- inst:num.examples
    }
    else{
      matrix.examples <- online.dataset[inst:(inst+(step-1)),,drop=F]
      rownames(matrix.examples) <- inst:(inst+(step-1))
    } 
    
    rownames.matrix.examples <- as.numeric(rownames(matrix.examples))
    num.examples.matrix.examples <- nrow(matrix.examples)
    
    # Num patterns (sets of neurons) to be considered in the chunck of examples (matrix.examples)
    num.patterns <- nrow(mapping$class.probabilities)
    
    # Get k nearest neighbour neurons for the set of examples
    n.k <- ceiling(mapping$z)
    if(n.k %% 2 == 0){n.k <- n.k + 1} # Guarantee odd number
    if(nrow(mapping$som.micro.clusters[[1]][[1]]$codes[[1]]) < n.k){
      # We cannot k less than sample size (number of neurons)
      n.k <- nrow(mapping$som.micro.clusters[[1]][[1]]$codes[[1]]) 
    }  
    winner <- get.knnx(mapping$som.micro.clusters[[1]][[1]]$codes[[1]],matrix.examples,k=n.k)
    
    # For each instance
    for(inst.l in 1:num.examples.matrix.examples){
      
      # Time stamp
      time.stamp <- time.stamp+1
      
      # Get incoming example
      x <- as.numeric(matrix.examples[inst.l,])
      
      # Create vector of possible predictions
      pred <- vector(length=nrow(mapping$class.probabilities), mode="numeric")
      
      # Controls if the instance is explained by the model
      explained <- 0
      
      # Check output of the winning neuron
      radiusFactor1 <- mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,1]]][[9]]
      radiusFactor2 <- mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,1]]][[12]]
      
      if(novel.classes[1] == 0){
        radiusFactor1 <- 9999999999999999999999
        radiusFactor2 <- 9999999999999999999999
      }   
      cat(rownames.matrix.examples[inst.l]," - ",exp(-winner$nn.dist[inst.l,1])," - ",winner$nn.dist[inst.l,1]," - ", radiusFactor1," - ",radiusFactor2,"\n")
      
      # Check if instance is explained by model checking the winner neuron
      if(winner$nn.dist[inst.l,1] <= radiusFactor1){
        
        # Instance was explained! 
        totally.explained <- 1  
        explained <- 1
        
        # Ceiling label cardinality 
        z <- ceiling(mapping$z)
        #z <- floor(mapping$z)
        
        # z is min(z,number of neurons)
        z <- min(z,length(winner$nn.index))
        for(j in 1:z){
          
          # Prototype vector of the neuron
          prototype.vector <- mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[5]]
          
          # Average output of the neuron
          sumOutputs <- mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[11]][1]
          numExamples <- mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[11]][2]
          average.output <- sumOutputs / numExamples

          if(j == 1){
            mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[11]][1] <- sumOutputs + exp(-winner$nn.dist[inst.l,j])
            mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[11]][2] <- numExamples + 1
            #pred[which(prototype.vector > 0)] <- 1
            
            # Classify in the max class
            id.max <- which.max(prototype.vector)
            pred[id.max] <- 1
            
            # Try the other classes, sorting the indexes of the prototype vector
            # according to the max class in it
            aux.classes.k <- sort(prototype.vector, index.return=T, decreasing = T)
            classes.k.1 <- which(aux.classes.k$x > 0)
            classes.k <- aux.classes.k$ix[classes.k.1]
            
            #classes.k <- which(prototype.vector > 0)
            indexes.idx.j <- classes.k[classes.k != id.max]
            
            for(idx.j in indexes.idx.j){
              prob.j <- mapping$class.probabilities[idx.j,idx.j] # p(y_j)
              prob.x.j <- exp(-winner$nn.dist[inst.l,j]) # p(x|y_j)
              prob.k.j <- 1
              
              for(idx.k in classes.k){
                if(pred[idx.k] == 1 & idx.k != idx.j){
                  prob.k.j <- prob.k.j * mapping$class.probabilities[idx.k,idx.j]
                }
              }
              prob.j.ks.x <- prob.j * prob.k.j * prob.x.j #p(y_j | y_k, x)
              
              # Threshold for neuron winner$nn.index[inst.l,j] considering class idx.j               
              cond.prob.threshold <- mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[10]][idx.j]
              browser()
              
              # Test for threshold!
              if(prob.j.ks.x >= cond.prob.threshold){
                
                sumOutputs <- mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[11]][1]
                numExamples <- mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[11]][2]
                average.output <- sumOutputs / numExamples
                
                mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[11]][1] <- sumOutputs + exp(-winner$nn.dist[inst.l,j])
                mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[11]][2] <- numExamples + 1
                pred[idx.j] <- 1
              }
            } # End for(idx.j in indexes.idx.j){
            
            # Maybe there is a concept evolution here (extra class) or a concept drift 
            if(winner$nn.dist[inst.l,j] > radiusFactor2){
              totally.explained <- 0
            }
          }
          else{
            # Next closest neuron
            
            # Sorting the indexes of the prototype vector
            # according to the max class in it
            aux.classes.j <- sort(prototype.vector, index.return=T, decreasing = T)
            classes.j.1 <- which(aux.classes.j$x > 0)
            indexes.idx.j <- aux.classes.j$ix[classes.j.1]
            
            #indexes.idx.j <- which(prototype.vector > 0)
            
            #if(sum(prototype.vector) > 1){
            #  browser() 
            #}
            
            for(idx.j in indexes.idx.j){
              if(pred[idx.j] == 0){ # Calculations are needed only if class was not predicted yet
                prob.j <- mapping$class.probabilities[idx.j,idx.j] # p(y_j)
                prob.x.j <- exp(-winner$nn.dist[inst.l,j]) # p(x|y_j)
                prob.k.j <- 1
                for(k in 1:(j-1)){ # p(y_k|y_j)
                  prototype.vector.k <- mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,k]]][[5]]
                
                  # Sorting the indexes of the prototype vector
                  # according to the max class in it
                  aux.classes.k <- sort(prototype.vector.k, index.return=T, decreasing = T)
                  classes.k.1 <- which(aux.classes.k$x > 0)
                  classes.k <- aux.classes.k$ix[classes.k.1] # prototype vector of the previous neuron in the rank of winners
                
                  #classes.k <- which(prototype.vector.k > 0) # prototype vector of the previous neuron in the rank of winners

                  for(idx.k in classes.k){
                    if(pred[idx.k] == 1 & idx.k != idx.j){
                      prob.k.j <- prob.k.j * mapping$class.probabilities[idx.k,idx.j]
                    }
                  } 
                }
                prob.j.ks.x <- prob.j * prob.k.j * prob.x.j #p(y_j | y_k, x)
               
                # Threshold for neuron winner$nn.index[inst.l,j] considering class idx.j               
                cond.prob.threshold <- mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[10]][idx.j]
              
                # Test for threshold!
                if(prob.j.ks.x >= cond.prob.threshold){
                
                  sumOutputs <- mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[11]][1]
                  numExamples <- mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[11]][2]
                  average.output <- sumOutputs / numExamples
                
                  mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[11]][1] <- sumOutputs + exp(-winner$nn.dist[inst.l,j])
                  mapping$som.micro.clusters[[1]][[2]][[winner$nn.index[inst.l,j]]][[11]][2] <- numExamples + 1
                
                  pred[idx.j] <- 1
                }
              } # End if(pred[idx.j] == 0)  
            } # End for(idx.j in indexes.idx.j){
          }
        } # End for(j in 1:z)
        
        # If totally.explained == 0, it means that for the winning neuron that recognized the example
        # the distance from the centroid was higher than radius factor 2
        # Thus the example can also belongs to a novel class
        if(totally.explained == 0){
          indexes.unknown.classified <- c(indexes.unknown.classified,rownames.matrix.examples[inst.l])
        }
      } # End if(winner$nn.dist[inst.l,1] <= radiusFactor1)
      
      if(explained == 1){ # Model explained the example
        
        # Store index		
        indexes.explained <- c(indexes.explained,rownames.matrix.examples[inst.l])
        
        # Update information about the model
        if(update.model.info == 1){
          mapping <- update.model.information(mapping,pred,x,time.stamp,n0,initial.number.classes,predictions,winner,inst.l)
        }   
        
        # Include predictions in the prediction matrix
        predictions <- rbind(predictions,pred)
        rownames(predictions)[nrow(predictions)] <- rownames.matrix.examples[inst.l]
        
        # Update class totals and class probabilities
        pred <- matrix(pred, nrow = 1, ncol = length(pred))
        mapping <- updateClassTotalsProbabilities(mapping,pred,1,predictions,initial.number.classes,0,num.offline.instances)
        
        # Update label cardinality
        mapping$z <- ( (mapping$total.instances - 1) * mapping$z + sum(apply(pred,1,sum)) ) / mapping$total.instances
        
        # Update conditional probabilitiy thresholds
        mapping$som.micro.clusters[[1]][[2]] <- updateCondProbabilitiesNeurons(mapping$som.micro.clusters[[1]][[2]],mapping$class.probabilities)
      }
    } # End for(inst.l in 1:num.examples.matrix.examples)
  } # End for(inst in seq(from=1, to=num.examples, by=step))

  cat("\n") 
  
  results$predictions <- predictions
  results$indexes.explained <- indexes.explained
  results$indexes.unknown.classified <- indexes.unknown.classified
  results$indexes.unknown.unclassified <- indexes.unknown.unclassified
  results$all.unclassified.indexes <- all.unclassified.indexes
  results$time.stamp.stm.removed <- time.stamp.stm.removed
  results$mapping <- mapping
  
  return(results)
}
# -------------------------------------------------------------------------------------