#####################################################################
# Implementation of a global version for the MLSC Kohonen map       #
# Ricardo Cerri - 10/10/2022                                        #
#####################################################################

# Auxiliary functions
source("functions.R")
source("kohonen.R")

# Package to work with multi-label data (mldr_0.4.2)
library(mldr)

# Package for the Kohonen neural network (kohonen_3.0.8)
library(kohonen)

# Package for get.knn function (FNN_1.1.3 )
library(FNN)

# Command line argument (parameter file)
args <- commandArgs(trailingOnly = TRUE)

# Read parameter values
parameters <- getParameterValues(args[1])
#parameters <- getParameterValues("MOA-5C-7C-2D-conf.txt")
#parameters <- getParameterValues("mediamill-conf.txt")

# Load datasets
mldrDataTrain <- mldr(parameters$train.data)
mldrDataTest <- mldr(parameters$test.data)

# Get data information
mldrDataTrain.labels.index <- mldrDataTrain$labels$index
mldrDataTrain.dataset <- mldrDataTrain$dataset
mldrDataTrain.attributesIndexes <- mldrDataTrain$attributesIndexes
num.labels <- nrow(mldrDataTrain$labels)
num.classes <- parameters$num.classes # Number of known classes
novel.classes <- parameters$novel.classes # The novelty classes

# Test Data
mldrDataTest.dataset <- mldrDataTest$dataset
mldrDataTest.attributesIndexes <- mldrDataTest$attributesIndexes
mldrDataTest.labels.index <- mldrDataTest$labels$index

# Get offline dataset
offline.dataset <- mldrDataTrain.dataset[,mldrDataTrain.attributesIndexes]
offline.dataset <- as.matrix(offline.dataset)
if(novel.classes[1] != 0){
  offline.classes <- mldrDataTrain.dataset[,mldrDataTrain.labels.index[-novel.classes]]
  online.classes <- mldrDataTest.dataset[,mldrDataTest.labels.index[-novel.classes]]
}else{
  offline.classes <- mldrDataTrain.dataset[,mldrDataTrain.labels.index]
  online.classes <- mldrDataTest.dataset[,mldrDataTest.labels.index]
}

# Get kohonen map information and parameters
n0 <- parameters$n0
n1 <- parameters$n1
tr.mode <- parameters$train.mode
num.iterations <- parameters$num.iterations
om <- parameters$omega
th <- parameters$theta
min.ex <- parameters$min.examples.cluster
window.stm.check <- parameters$window.stm.check
empty.predictions <- parameters$empty.predictions
n.neurons.grid <- parameters$grid.dimension
update.model.info <- parameters$update.model.info

# Number of instances for the offline phase
num.offline.instances <- nrow(offline.dataset)

# Execute only once?
only.once <- 0

# Run experiments for different number of neurons (until one neuron per class)
repeat{
  
  # Store results of all executions
  maps <- list()
  vet <- vector()
  vet.NP <- vector()
  
  # Repeat the experiment num.exp times
  executed <- 1
  for(num.exp in 1:parameters$num.experiments){
    
    cat("\nExperiment ",num.exp)
    
    # Build kohonen map in the offile examples
    res <- try(mapping <- kohonen.offline.global(offline.dataset,offline.classes,num.iterations,n0,n1,n.neurons.grid,tr.mode,min.ex))
    if(inherits(res, "try-error")){
      executed <- 0
      break
    }
    
    # Test to see the performance of the network in the mapped instances
    #offline.predictions <- predict.kohonen.win.bayes(mapping$som.micro.clusters,offline.dataset,mapping$class.probabilities,mapping$z)
    #results.offline <- macro.precision.recall.fmeasure(offline.classes,offline.predictions)
    
    # Beginning of the online phase
    
    # Examples without classes
    online.dataset <- mldrDataTest.dataset[,mldrDataTest.attributesIndexes]
    
    # Get online dataset
    online.dataset <- as.matrix(online.dataset)
    
    # Verify the examples one by one
    #th <- mapping$theta
    #min.ex <- mapping$min.instances.neuron
    th <- parameters$theta
    min.ex <- parameters$min.examples.cluster
    
    window.stm.check <- th
    results <- kohonen.online.bayes.ND(mapping,online.dataset,n0,n1,tr.mode,num.iterations,om,
                                       th,min.ex,window.stm.check,novel.classes,n.neurons.grid,
                                       update.model.info,num.offline.instances)
    
    # Evaluate the online predictions. Only considering the known classes
    true.classes <- online.classes[results$indexes.explained,]
    num.evaluation.windows <- parameters$num.evaluation.windows
    
    # Should we consider the unknown examples as empty predictions?
    if(empty.predictions == 1){
      
      empty.true <- online.classes[results$all.unclassified.indexes,]
      true.classes <- rbind(true.classes,empty.true)
      if(novel.classes[1] != 0){
        empty.pred <- matrix(nrow=length(results$all.unclassified.indexes), ncol=length(mldrDataTest.labels.index[-novel.classes]),data=0)
      }else{
        empty.pred <- matrix(nrow=length(results$all.unclassified.indexes), ncol=length(mldrDataTest.labels.index),data=0)
      }
      rownames(empty.pred) <- results$all.unclassified.indexes
      results$predictions <- rbind(results$predictions,empty.pred)
    }
    
    results.online <- macro.precision.recall.fmeasure.windows(true.classes,results$predictions,num.evaluation.windows)
    results$true.classes <- true.classes
    results$results.online <- results.online
    
    # Do we have novel classes?
    if(novel.classes[1] != 0){
      # Evaluate the online predictions. Considering known classes and NP
      online.classes.NP <- mldrDataTest.dataset[,mldrDataTest.labels.index]
      # Should we consider the unknown examples as empty predictions?
      if(empty.predictions == 1){
        true.classes.NP <- online.classes.NP[c(results$indexes.explained,results$all.unclassified.indexes),]
      }
      else{
        true.classes.NP <- online.classes.NP[results$indexes.explained,]
      }
      
      # Recover predictions with NPs associated with new classes
      results.association <- associate.NP.classes(results$predictions,true.classes.NP,novel.classes,num.evaluation.windows,results$mapping$NP)
      predictions.NP <- results.association$predictions.NP
      associations <- results.association$associations
      results$associations <- associations
      
      # Evaluation
      online.classes.NP.only <- mldrDataTest.dataset[,mldrDataTest.labels.index[novel.classes]]
      # Should we consider the unknown examples as empty predictions?
      if(empty.predictions == 1){
        true.classes.NP.only <- online.classes.NP.only[c(results$indexes.explained,results$all.unclassified.indexes),]
      }
      else{
        true.classes.NP.only <- online.classes.NP.only[results$indexes.explained,]
      }
      true.classes.NP <- cbind(true.classes,true.classes.NP.only)
      
      # Precision and Recall per window
      results.online.NP <- macro.precision.recall.fmeasure.windows.NP(true.classes.NP,predictions.NP,num.evaluation.windows,novel.classes)
      results$predictions.NP <- predictions.NP
      results$true.classes.NP <- true.classes.NP
      results$results.online.NP <- results.online.NP
      
      # Unknown rate calculation per window
      true.classes.for.unk <- mldrDataTest.dataset[,c(mldrDataTest.labels.index[-novel.classes],mldrDataTest.labels.index[novel.classes])]
      unknown.rate <- unknown.rate.windows(results$all.unclassified.indexes,results$time.stamp.stm.removed,true.classes.for.unk,num.evaluation.windows,novel.classes)
      results$unknown.rate.window <- unknown.rate$unknown.rate.window
      results$num.unknown.window <- unknown.rate$num.unknown.window
    } # End if(novel.classes[1] != 0)
    
    # Save results of all executions
    maps[[num.exp]] <- results
    vet <- c(vet,results.online$ma.fmeasure)
    if(novel.classes[1] != 0){
      vet.NP <- c(vet.NP,results.online.NP$ma.fmeasure)
    }
  } # End for(num.exp in 1:parameters$num.experiments)
  
  # Save the predictions, parameter values and model
  if(executed == 1){
    name.dataset <- unlist(strsplit(parameters$test.data,"/"))
    name.dataset <- gsub(" ","",name.dataset[length(name.dataset)])
    time <- format(Sys.time(), "%H.%M.%OS")
    
    capture.output(parameters,file=paste("Results/",name.dataset,"-",time,"-parameters-",n.neurons.grid,".txt",sep=""))
    saveRDS(maps, file=paste("Results/",name.dataset,"-",time,"-model-",n.neurons.grid,".rds",sep=""))
    capture.output(
      cat("Dataset: ",name.dataset,", Mean: ",mean(vet),", StdDev: ",sd(vet),"\n",
          "Dataset: ",name.dataset,", Mean NP: ",mean(vet.NP),", StdDev: ",sd(vet.NP)),
      file=paste("Results/",name.dataset,"-",time,"-mean.fmeasures-",n.neurons.grid,".txt",sep="")
    )
    #readRDS
  } # End if(executed == 1)
  
  if(only.once == 1){
    break
  }
  
  # Reduce dimension of the grid by one neuron
  n.neurons.grid <- n.neurons.grid-1
  
  # If grid with 1 neurons, stop
  if(n.neurons.grid == 1){
    break
  }
}
# -------------------------------------------------------------------------------------



